@@ -1,104 +1,193 @@
 use std::{
-    convert::TryInto,
+    cell::UnsafeCell,
+    fmt::{self, Debug},
     marker::PhantomData,
-    mem::{forget, transmute},
+    mem::transmute,
 };
 
 use crate::{
     environment::{params_to_vec, Param},
     error::LunaticError,
     host_api::{self, message, process},
-    mailbox::{LinkMailbox, Mailbox, TransformMailbox},
-    message::Message,
+    mailbox::{LinkMailbox, Mailbox, MessageRw, TransformMailbox},
+    request::Request,
+    tag::Tag,
+    Environment,
+};
+
+use rmp_serde::decode;
+use serde::{
+    de::{self, DeserializeOwned, Visitor},
+    Deserialize, Deserializer, Serialize, Serializer,
 };
 
 /// A sandboxed computation.
 ///
 /// Processes are fundamental building blocks of Lunatic applications. Each of them has their own
-/// memory space. The only way for processes to interact is trough [`Message`] passing.
+/// memory space. The only way for processes to interact is trough [`Serialize + DeserializeOwned`]
+/// passing.
 ///
 /// ### Safety:
 /// It's not safe to use mutable `static` variables to share data between processes, because each
 /// of them is going to see a separate heap and a unique `static` variable.
-pub struct Process<T: Message> {
-    id: u64,
+pub struct Process<T: Serialize + DeserializeOwned> {
+    pub(crate) id: u64,
+    // If the process handle is serialized it will be removed from our resources, so we can't call
+    // `drop_process()` anymore on it.
+    consumed: UnsafeCell<bool>,
     _phantom: PhantomData<T>,
 }
 
-impl<T: Message> Clone for Process<T> {
+impl<T: Serialize + DeserializeOwned> Debug for Process<T> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        let mut uuid: [u8; 16] = [0; 16];
+        unsafe { host_api::process::id(self.id, &mut uuid as *mut [u8; 16]) };
+        f.debug_struct("Process")
+            .field("uuid", &u128::from_le_bytes(uuid))
+            .finish()
+    }
+}
+
+impl<T: Serialize + DeserializeOwned> Clone for Process<T> {
     fn clone(&self) -> Self {
         let id = unsafe { host_api::process::clone_process(self.id) };
         Process::from(id)
     }
 }
 
-impl<T: Message> Drop for Process<T> {
+impl<T: Serialize + DeserializeOwned> Drop for Process<T> {
     fn drop(&mut self) {
-        unsafe { process::drop_process(self.id) };
+        // Only drop process if it's not already consumed
+        if unsafe { !*self.consumed.get() } {
+            unsafe { process::drop_process(self.id) };
+        }
+    }
+}
+impl<T: Serialize + DeserializeOwned> Serialize for Process<T> {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        // Mark process as consumed
+        unsafe { *self.consumed.get() = true };
+
+        let index = unsafe { host_api::message::push_process(self.id) };
+        serializer.serialize_u64(index)
+    }
+}
+struct ProcessVisitor<T> {
+    _phantom: PhantomData<T>,
+}
+impl<'de, T: Serialize + DeserializeOwned> Visitor<'de> for ProcessVisitor<T> {
+    type Value = Process<T>;
+
+    fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
+        formatter.write_str("an u64 index")
+    }
+
+    fn visit_u64<E>(self, index: u64) -> Result<Self::Value, E>
+    where
+        E: de::Error,
+    {
+        let id = unsafe { host_api::message::take_process(index) };
+        Ok(Process::from(id))
     }
 }
 
-impl<T: Message> Message for Process<T> {
-    fn from_bincode(data: &[u8], resources: &[u64]) -> (usize, Self) {
-        // The serialized value for a process is the u64 index inside the resources array.
-        // The resources array will contain the new resource index.
-        let index = u64::from_le_bytes(data.try_into().unwrap());
-        let proc = Process::from(resources[index as usize]);
-        (8, proc)
-    }
-
-    #[allow(clippy::wrong_self_convention)]
-    unsafe fn to_bincode(self, dest: &mut Vec<u8>) {
-        let index = message::add_process(self.id);
-        dest.extend(index.to_le_bytes());
-        // By adding the process to the message it will be removed from our resources.
-        // Dropping it would cause a trap.
-        forget(self);
+impl<'de, T: Serialize + DeserializeOwned> Deserialize<'de> for Process<T> {
+    fn deserialize<D>(deserializer: D) -> Result<Process<T>, D::Error>
+    where
+        D: Deserializer<'de>,
+    {
+        deserializer.deserialize_u64(ProcessVisitor {
+            _phantom: PhantomData {},
+        })
     }
 }
 
-impl<T: Message> Process<T> {
+impl<T: Serialize + DeserializeOwned> Process<T> {
     pub(crate) fn from(id: u64) -> Self {
         Process {
             id,
+            consumed: UnsafeCell::new(false),
             _phantom: PhantomData,
         }
     }
 
+    /// Send message to process.
     pub fn send(&self, value: T) {
+        self.send_(None, value)
+    }
+
+    /// Tag a message and send it to a process.
+    pub(crate) fn tag_send(&self, tag: Tag, value: T) {
+        self.send_(Some(tag.0), value)
+    }
+
+    fn send_(&self, tag: Option<i64>, value: T) {
+        let tag = tag.unwrap_or(0);
         // Create new message
-        unsafe { message::create() };
-        let mut buffer = Vec::new();
-        unsafe { value.to_bincode(&mut buffer) };
-        // Set sending buffer
-        unsafe { message::set_buffer(buffer.as_ptr(), buffer.len()) };
-        // During bincode serialization resources will add themself to the message
+        unsafe { message::create_data(tag, 0) };
+        // During serialization resources will add themself to the message
+        rmp_serde::encode::write(&mut MessageRw {}, &value).unwrap();
         // Send it
         unsafe { message::send(self.id) };
     }
 
-    #[allow(clippy::result_unit_err)]
-    pub fn join(&self) -> Result<(), ()> {
-        match unsafe { process::join(self.id) } {
-            0 => Ok(()),
-            _ => Err(()),
-        }
+    /// Links the current process with another one.
+    pub fn link(&self) {
+        unsafe { process::link(0, self.id) };
+    }
+
+    /// Unlinks the current process from another one.
+    pub fn unlink(&self) {
+        unsafe { process::unlink(self.id) };
+    }
+}
+
+impl<T, U> Process<Request<T, U>>
+where
+    T: Serialize + DeserializeOwned,
+    U: Serialize + DeserializeOwned,
+{
+    pub fn request(&self, message: T) -> Result<U, decode::Error> {
+        // The response can be an arbitrary type and doesn't need to match the the current one.
+        let one_time_mailbox = unsafe { Mailbox::<U>::new() };
+        let sender_process = this(&one_time_mailbox);
+        let tag = Tag::new();
+        let request = Request::new(message, tag, sender_process);
+        // Create new message
+        unsafe { message::create_data(tag.0, 0) };
+        // During serialization resources will add themself to the message
+        rmp_serde::encode::write(&mut MessageRw {}, &request).unwrap();
+        // Send it and wait for an replay
+        unsafe { message::send_receive_skip_search(self.id, 0) };
+        // Read the message out from the scratch buffer
+        rmp_serde::from_read(MessageRw {})
     }
 }
 
 /// Returns a handle to the current process.
-pub fn this<T: Message, U: TransformMailbox<T>>(mailbox: U) -> (Process<T>, U) {
+pub fn this<T: Serialize + DeserializeOwned, U: TransformMailbox<T>>(_mailbox: &U) -> Process<T> {
     let id = unsafe { process::this() };
-    (Process::from(id), mailbox)
+    Process::from(id)
+}
+
+/// Returns a handle to the current environment.
+pub fn this_env() -> Environment {
+    let id = unsafe { process::this_env() };
+    Environment::from(id)
 }
 
 /// Spawns a new process from a function.
 ///
 /// - `function` is the starting point of the new process. The new process doesn't share
 ///   memory with its parent, because of this the function can't capture anything from parents.
-pub fn spawn<T: Message>(function: fn(Mailbox<T>)) -> Result<Process<T>, LunaticError> {
+pub fn spawn<T: Serialize + DeserializeOwned>(
+    function: fn(Mailbox<T>),
+) -> Result<Process<T>, LunaticError> {
     // LinkMailbox<T> & Mailbox<T> are marker types and it's safe to cast to Mailbox<T> here if we
-    //  set the `link` argument to `false`.
+    // set the `link` argument to `false`.
     let function = unsafe { transmute(function) };
     spawn_(None, false, Context::<(), _>::Without(function))
 }
@@ -112,11 +201,11 @@ pub fn spawn_link<T, P, M>(
     function: fn(Mailbox<T>),
 ) -> Result<(Process<T>, LinkMailbox<P>), LunaticError>
 where
-    T: Message,
-    P: Message,
+    T: Serialize + DeserializeOwned,
+    P: Serialize + DeserializeOwned,
     M: TransformMailbox<P>,
 {
-    let mailbox = mailbox.catch_child_panic();
+    let mailbox = mailbox.catch_link_panic();
     let proc = spawn_(None, true, Context::<(), _>::Without(function))?;
     Ok((proc, mailbox))
 }
@@ -132,11 +221,11 @@ pub fn spawn_link_unwrap<T, P, M>(
     function: fn(Mailbox<T>),
 ) -> Result<(Process<T>, Mailbox<P>), LunaticError>
 where
-    T: Message,
-    P: Message,
+    T: Serialize + DeserializeOwned,
+    P: Serialize + DeserializeOwned,
     M: TransformMailbox<P>,
 {
-    let mailbox = mailbox.panic_if_child_panics();
+    let mailbox = mailbox.panic_if_link_panics();
     let proc = spawn_(None, true, Context::<(), _>::Without(function))?;
     Ok((proc, mailbox))
 }
@@ -144,12 +233,12 @@ where
 /// Spawns a new process from a function and context.
 ///
 /// - `context` is  data that we want to pass to the newly spawned process. It needs to impl.
-///    the [`Message`] trait.
+///    the [`Serialize + DeserializeOwned`] trait.
 ///
 /// - `function` is the starting point of the new process. The new process doesn't share
 ///   memory with its parent, because of this the function can't capture anything from parents.
 ///   The first argument of this function is going to be the received `context`.
-pub fn spawn_with<C: Message, T: Message>(
+pub fn spawn_with<C: Serialize + DeserializeOwned, T: Serialize + DeserializeOwned>(
     context: C,
     function: fn(C, Mailbox<T>),
 ) -> Result<Process<T>, LunaticError> {
@@ -162,7 +251,7 @@ pub fn spawn_with<C: Message, T: Message>(
 /// Spawns a new process from a function and context, and links it to the parent.
 ///
 /// - `context` is  data that we want to pass to the newly spawned process. It needs to impl.
-///    the [`Message`] trait.
+///    the [`Serialize + DeserializeOwned`] trait.
 ///
 /// - `function` is the starting point of the new process. The new process doesn't share
 ///   memory with its parent, because of this the function can't capture anything from parents.
@@ -173,12 +262,12 @@ pub fn spawn_link_with<C, T, P, M>(
     function: fn(C, Mailbox<T>),
 ) -> Result<(Process<T>, LinkMailbox<P>), LunaticError>
 where
-    C: Message,
-    T: Message,
-    P: Message,
+    C: Serialize + DeserializeOwned,
+    T: Serialize + DeserializeOwned,
+    P: Serialize + DeserializeOwned,
     M: TransformMailbox<P>,
 {
-    let mailbox = mailbox.catch_child_panic();
+    let mailbox = mailbox.catch_link_panic();
     let proc = spawn_(None, true, Context::With(function, context))?;
     Ok((proc, mailbox))
 }
@@ -186,7 +275,7 @@ where
 /// Spawns a new process from a function and context, and links it to the parent.
 ///
 /// - `context` is  data that we want to pass to the newly spawned process. It needs to impl.
-///    the [`Message`] trait.
+///    the [`Serialize + DeserializeOwned`] trait.
 ///
 /// - `function` is the starting point of the new process. The new process doesn't share
 ///   memory with its parent, because of this the function can't capture anything from parents.
@@ -199,23 +288,23 @@ pub fn spawn_link_unwrap_with<C, T, P, M>(
     function: fn(C, Mailbox<T>),
 ) -> Result<(Process<T>, Mailbox<P>), LunaticError>
 where
-    C: Message,
-    T: Message,
-    P: Message,
+    C: Serialize + DeserializeOwned,
+    T: Serialize + DeserializeOwned,
+    P: Serialize + DeserializeOwned,
     M: TransformMailbox<P>,
 {
-    let mailbox = mailbox.panic_if_child_panics();
+    let mailbox = mailbox.panic_if_link_panics();
     let proc = spawn_(None, true, Context::With(function, context))?;
     Ok((proc, mailbox))
 }
 
-pub(crate) enum Context<C: Message, T: Message> {
+pub(crate) enum Context<C: Serialize + DeserializeOwned, T: Serialize + DeserializeOwned> {
     With(fn(C, Mailbox<T>), C),
     Without(fn(Mailbox<T>)),
 }
 
 // If `module_id` is None it will use the current module & environment.
-pub(crate) fn spawn_<C: Message, T: Message>(
+pub(crate) fn spawn_<C: Serialize + DeserializeOwned, T: Serialize + DeserializeOwned>(
     module_id: Option<u64>,
     link: bool,
     context: Context<C, T>,
@@ -265,6 +354,7 @@ pub(crate) fn spawn_<C: Message, T: Message>(
             Context::With(_, context) => {
                 let self_ = Process {
                     id,
+                    consumed: UnsafeCell::new(false),
                     _phantom: PhantomData,
                 };
                 self_.send(context);
@@ -274,6 +364,7 @@ pub(crate) fn spawn_<C: Message, T: Message>(
             }
             Context::Without(_) => Ok(Process {
                 id,
+                consumed: UnsafeCell::new(false),
                 _phantom: PhantomData,
             }),
         }
@@ -288,16 +379,18 @@ pub fn sleep(milliseconds: u64) {
 }
 
 // Type helper
-fn type_helper_wrapper<T: Message>(function: usize) {
-    let mailbox = Mailbox::new();
+fn type_helper_wrapper<T: Serialize + DeserializeOwned>(function: usize) {
+    let mailbox = unsafe { Mailbox::new() };
     let function: fn(Mailbox<T>) = unsafe { transmute(function) };
     function(mailbox);
 }
 
 // Type helper with context
-fn type_helper_wrapper_context<C: Message, T: Message>(function: usize) {
-    let context = Mailbox::new().receive();
-    let mailbox = Mailbox::new();
+fn type_helper_wrapper_context<C: Serialize + DeserializeOwned, T: Serialize + DeserializeOwned>(
+    function: usize,
+) {
+    let context = unsafe { Mailbox::new() }.receive().unwrap();
+    let mailbox = unsafe { Mailbox::new() };
     let function: fn(C, Mailbox<T>) = unsafe { transmute(function) };
     function(context, mailbox);
 }
